"""
    awss3.py

    Copyright (c) 2016 Snowplow Analytics Ltd. All rights reserved.

    This program is licensed to you under the Apache License Version 2.0,
    and you may not use this file except in compliance with the Apache License
    Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
    http://www.apache.org/licenses/LICENSE-2.0.

    Unless required by applicable law or agreed to in writing,
    software distributed under the Apache License Version 2.0 is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
    express or implied. See the Apache License Version 2.0 for the specific
    language governing permissions and limitations there under.

    Authors: Joshua Beemster
    Copyright: Copyright (c) 2016 Snowplow Analytics Ltd
    License: Apache License Version 2.0
"""


from __future__ import division

import tempfile

import boto3
from boto3.s3.transfer import S3Transfer
from botocore.exceptions import ClientError

import release_manager.logger as logger
import release_manager.package as pack


class S3Location(object):
    """Entity storing information about where artifact should be uploaded.
    Parsed from package, by `get_locations`"""
    def __init__(self, bucket, path, region):
        self.bucket = bucket
        self.path = path
        self.region = region

    def __str__(self):
        return "AWS S3 bucket [{}] at [{}] region. Key [{}]".format(self.bucket, self.region, self.path)


def upload_to_s3(args, local, package, target):
    """Upload artifact to AWS S3"""

    if args.make_artifact:

        # Run build commands
        if "build_commands" in package.keys():
            pack.execute_commands(package["build_commands"])

        for artifact in package["artifacts"]:
            artifact_file = pack.create_artifact(
                local['root_dir'],
                package['version'],
                package['name'],
                artifact['type'],
                artifact['prefix'],
                artifact['suffix'],
                artifact['binary_paths']
            )

            # Process all S3 locations
            for location in get_locations(package):
                full_s3_path = get_full_path(location, artifact_file)

                client = boto3.client('s3', region_name=location.region, aws_access_key_id=target['access_key_id'], aws_secret_access_key=target['secret_access_key'])
                transfer = S3Transfer(client)

                if package['override']:
                    transfer.upload_file(artifact_file['artifact_path'], location.bucket, full_s3_path)
                else:
                    try:
                        transfer.download_file(location.bucket, full_s3_path, tempfile.mktemp())
                        if package['continue_on_conflict']:
                            logger.log_info("Artifact [%s] exists, but continue_on_conflict flag is true, not failing deploy..." % full_s3_path)
                        else:
                            raise ValueError("Artifact at %s already exists" % (full_s3_path,))
                    except ClientError as e:
                        if e.response['Error']['Code'] == '404':
                            transfer.upload_file(artifact_file['artifact_path'], location.bucket, full_s3_path)
                            logger.log_info("Artifact uploaded to {}".format(location))
                        else:
                            logger.log_info("continue_on_conflict flag is true, not failing release...")
                            raise

    else:
        logger.log_info("make-artifact flag was not passed. Do nothing")


def get_locations(package):
    """Return list of S3 locations extracted from package. Always return array"""
    if 'bucket' in package and 'locations' in package:
        raise RuntimeError("Package cannot contain both 'bucket' and 'locations' property")
    elif 'bucket' in package:
        try:
            return [S3Location(package['bucket'], package['path'], package['region'])]
        except KeyError:
            raise RuntimeError("Package missing required keys (bucket, path, region)")
    elif 'locations' in package:
        return [S3Location(location['bucket'], location['path'], location['region']) for location in package['locations']]
    else:
        raise RuntimeError("Either 'locations' array or 'bucket' must be present")


def get_full_path(location, artifact_file):
    if location.path.endswith('/'):
        path = location.path
    elif not location:
        path = ''
    else:
        path = location.path + '/'
    return path + artifact_file['artifact_name']



"""
    utils.py

    Copyright (c) 2016 Snowplow Analytics Ltd. All rights reserved.

    This program is licensed to you under the Apache License Version 2.0,
    and you may not use this file except in compliance with the Apache License
    Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
    http://www.apache.org/licenses/LICENSE-2.0.

    Unless required by applicable law or agreed to in writing,
    software distributed under the Apache License Version 2.0 is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
    express or implied. See the Apache License Version 2.0 for the specific
    language governing permissions and limitations there under.

    Authors: Joshua Beemster
    Copyright: Copyright (c) 2016 Snowplow Analytics Ltd
    License: Apache License Version 2.0
"""

import contextlib
import os
import subprocess
import random
import string
import threading
import re
import time

from jinja2 import Template
import yaml
from yaml.loader import SafeLoader

import release_manager.logger as logger


# --- Command Execution

@contextlib.contextmanager
def working_directory(path):
    """A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.

    """
    prev_cwd = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(prev_cwd)


def output_everything(output):
    """Callback to log stdout"""
    (stdout, stderr) = output.communicate()
    if output.returncode == 0:
        logger.log_output(stdout)
    else:
        logger.log_output("Process has failed.\n" + stdout.decode("utf-8"))
        raise ValueError(stderr)


def output_value(output, fail_on_err=False):
    """Returns the value of the command output"""
    (stdout, stderr) = output.communicate()
    if output.returncode == 0:
        return stdout
    else:
        if fail_on_err:
            raise ValueError(stderr)
        else:
            return stderr


def execute(command, callback=output_everything, quiet=False, shell=False):
    """Execute shell command with optional callback"""
    if not quiet:
        formatted_command = " ".join(command) if (type(command) == list) else command
        logger.log_info("Executing [{0}]".format(formatted_command))

    output = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell)

    if hasattr(callback, '__call__'):
        return callback(output)
    else:
        return output


class AwakeThread(threading.Thread):
    """Thread printing messages to console, keeping Travis CI from shut down prematurely"""
    def run(self):
        print("Starting AwakeThread")
        count = 0
        while count < 10:
            message = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(128))
            print("Tick %s. %s seconds passed. Random output: %s" % (str(count), str(40 * count), message))
            time.sleep(40)
            count = count + 1
        print("Exit AwakeThread")


def sbt_version(directory):
    """Return version of project in some directory"""
    t = AwakeThread()
    t.setDaemon(True)
    t.start()

    with working_directory(directory):
        sbt_output = execute(['sbt', '-no-colors', 'version'], None)
        print(sbt_output.stderr.read())
        for line in sbt_output.stdout.read().split("\n"):
            print(line)
            match = re.search('\[info\]\s*(\d+\.\d+\.\d+.*)$', line)
            if match:
                return match.group(1)
        raise ValueError("Not SBT Project: " + directory)

    t.join()
    print("Some result should appear")



PREDEFINED_FUNCTIONS = {
    'sbt_version': sbt_version
}

# --- Config

def parse_config(config_path):
    """Checks if secrets need to be fetched from the environment"""

    temp_path = "%s.tmp" % config_path

    # Template yaml file
    with open(config_path, 'r') as stream:
        try:
            temp = template_yaml(yaml.safe_load(stream))
        except yaml.YAMLError as exc:
            raise ValueError("Invalid config passed to the program: %s" % exc)

    # Dump templated file back to file-system
    with open(temp_path, 'w') as outfile:
        yaml.safe_dump(temp, outfile, default_flow_style=False)

    # Add environment resolver
    pattern_env = re.compile(r'^(.*)\<%= ENV\[\'(.*)\'\] %\>(.*)$')
    yaml.add_implicit_resolver("!pathex", pattern_env, None, SafeLoader)

    # Add command resolver
    pattern_cmd = re.compile(r'^(.*)\<%= CMD\[\'(.*)\'\] %\>(.*)$')
    yaml.add_implicit_resolver("!pathcmd", pattern_cmd, None, SafeLoader)

    # Add function resolver
    pattern_fun = re.compile(r'^(.*)\<%= FUNC\[\'(.*)\((.*)\)\'\] %\>(.*)$')
    yaml.add_implicit_resolver("!func", pattern_fun, None, SafeLoader)

    def pathex_constructor(loader, node):
        """Processes environment variables found in the YAML"""
        value = loader.construct_scalar(node)
        before_path, env_var, remaining_path = pattern_env.match(value).groups()
        return before_path + os.environ[env_var] + remaining_path

    def pathcmd_constructor(loader, node):
        """Processes command variables found in the YAML"""
        value = loader.construct_scalar(node)
        before_path, cmd_var, remaining_path = pattern_cmd.match(value).groups()
        retval = output_value(execute(cmd_var, None, True, True), True)
        return before_path + retval.decode("utf-8") + remaining_path

    def fun_constructor(loader, node):
        """Processes embedded functions found in the YAML"""
        value = loader.construct_scalar(node)
        before_path, fun, arg, remaining_path = pattern_fun.match(value).groups()
        retval = PREDEFINED_FUNCTIONS[fun](arg)
        return before_path + retval.decode("utf-8") + remaining_path

    yaml.add_constructor("!pathex", pathex_constructor, SafeLoader)
    yaml.add_constructor("!pathcmd", pathcmd_constructor, SafeLoader)
    yaml.add_constructor("!func", fun_constructor, SafeLoader)

    with open(temp_path, 'r') as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            raise ValueError("Invalid config passed to the program: %s" % exc)


def template_yaml(yaml_dict):
    """Runs the YAML through the Jinja2 templater"""
    template_yaml_dict = Template(yaml.safe_dump(yaml_dict, default_flow_style=False))
    template_rendered = template_yaml_dict.render(yaml_dict)
    return yaml.safe_load(template_rendered)



"""
    bintray.py

    Copyright (c) 2016 Snowplow Analytics Ltd. All rights reserved.

    This program is licensed to you under the Apache License Version 2.0,
    and you may not use this file except in compliance with the Apache License
    Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
    http://www.apache.org/licenses/LICENSE-2.0.

    Unless required by applicable law or agreed to in writing,
    software distributed under the Apache License Version 2.0 is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
    express or implied. See the Apache License Version 2.0 for the specific
    language governing permissions and limitations there under.

    Authors: Joshua Beemster
    Copyright: Copyright (c) 2016 Snowplow Analytics Ltd
    License: Apache License Version 2.0
"""


from __future__ import division

import json
import requests

import release_manager.logger as logger
import release_manager.package as pack


def create_bintray_version(version, package, repo, user_org, user, api_key):
    """Creates a new Bintray version for the package"""
    logger.log_start("Creating Bintray version %s in package %s" % (version, package))

    url = "https://api.bintray.com/packages/%s/%s/%s/versions" % (user_org, repo, package)

    payload = {
        'name': version,
        'desc': 'Release of %s' % package
    }
    headers = {
        'Content-Type': 'application/json'
    }

    response = requests.post(url, data=json.dumps(payload), headers=headers, auth=(user, api_key))

    code = response.status_code
    success = False

    if code == 409:
        logger.log_info("Bintray version %s already exists, skipping." % version)
        success = True
    else:
        code_family = code // 100
        if code_family == 2 or code_family == 3:
            logger.log_info("Bintray Version created!")
            success = True
        else:
            logger.log_info("Bintray API response %s is not 409 (package already exists) nor in 2xx or 3xx range" % code)
            success = False

    logger.log_done()
    return success


def upload_bintray_artifact(version, package, repo, user_org, user, api_key, artifact_name, artifact_path, publish, override, continue_on_conflict):
    """Uploads the artifact to Bintray"""
    logger.log_start("Uploading artifact to Bintray")

    url = 'https://api.bintray.com/content/%s/%s/%s/%s/%s' % (user_org, repo, package, version, artifact_name)
    parameters = {
        'publish': publish,
        'override': override
    }

    with open(artifact_path, "rb") as package_fp:
        response = requests.put(
            url,
            auth=(user, api_key),
            params=parameters,
            data=package_fp
        )

    code = response.status_code
    code_family = int(code) // 100
    success = False

    if code_family == 2 or code_family == 3:
        logger.log_info("Bintray artifact uploaded!")
        success = True
    else:
        logger.log_info("Bintray API response %s is not in 2xx or 3xx range" % code)
        if continue_on_conflict:
            logger.log_info("continue_on_conflict flag is true, not failing release...")
            success = True
        else:
            success = False

    logger.log_done()
    return success


def deploy_to_bintray(args, local, package, target):
    """Deploys the package to Bintray"""

    # Make the version
    if args.make_version:
        retval_1 = create_bintray_version(
            package["version"],
            package["name"],
            package["repo"],
            package["user_org"],
            target["user"],
            target["password"]
        )
        if retval_1 is False:
            raise ValueError("Could not create new Bintray version for the package!")

    # Make the artifacts
    if args.make_artifact:

        # Run build commands
        if "build_commands" in package.keys():
            pack.execute_commands(package["build_commands"])

        # Build artifacts
        for artifact in package["artifacts"]:
            retval_2 = pack.create_artifact(
                local["root_dir"],
                package["version"],
                package["name"],
                artifact["type"],
                artifact["prefix"],
                artifact["suffix"],
                artifact["binary_paths"]
            )

            # Upload the artifacts
            if args.make_artifact and args.upload_artifact:
                retval_3 = upload_bintray_artifact(
                    package["version"],
                    package["name"],
                    package["repo"],
                    package["user_org"],
                    target["user"],
                    target["password"],
                    retval_2["artifact_name"],
                    retval_2["artifact_path"],
                    "1" if package["publish"] else "0",
                    "1" if package["override"] else "0",
                    package["continue_on_conflict"]
                )
                if retval_3 is False:
                    raise ValueError("Could not upload artifact to Bintray!")



"""
    package.py

    Copyright (c) 2016 Snowplow Analytics Ltd. All rights reserved.

    This program is licensed to you under the Apache License Version 2.0,
    and you may not use this file except in compliance with the Apache License
    Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
    http://www.apache.org/licenses/LICENSE-2.0.

    Unless required by applicable law or agreed to in writing,
    software distributed under the Apache License Version 2.0 is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
    express or implied. See the Apache License Version 2.0 for the specific
    language governing permissions and limitations there under.

    Authors: Joshua Beemster
    Copyright: Copyright (c) 2016 Snowplow Analytics Ltd
    License: Apache License Version 2.0
"""


from __future__ import print_function
import os
import os.path

import release_manager.logger as logger
import release_manager.utils as utils


# --- Constants


ARTIFACT_STAGING_DIR = "dist"

AVAILABLE_ARTIFACT_TYPES = ['zip', 'asis']


# --- Functions


def create_zip_artifact(root_dir, version, package, artifact_prefix, artifact_suffix, binary_paths):
    """Builds the artifact for upload"""

    artifact_root = "%s%s%s" % (artifact_prefix, version, artifact_suffix,)
    artifact_name = "%s.zip" % artifact_root
    artifact_name = artifact_name.replace("-", "_")

    logger.log_info("Building artifact %s..." % artifact_name)

    artifact_folder = "%s/%s/%s" % (root_dir, ARTIFACT_STAGING_DIR, package)
    utils.execute(['mkdir', '-p', artifact_folder])

    file_names = []
    for path in binary_paths:
        utils.execute(['cp', path, artifact_folder])
        file_names.append(os.path.basename(path))

    os.chdir(artifact_folder)
    utils.execute(['zip', artifact_name] + file_names)
    os.chdir(root_dir)

    logger.log_done()

    return {
        'artifact_name': artifact_name,
        'artifact_path': "%s/%s" % (artifact_folder, artifact_name)
    }


def create_asis_artifact(root_dir, version, package, artifact_prefix, artifact_suffix, binary_paths):
    """Construct artifact name and perform no operations"""

    if len(binary_paths) != 1 and type(binary_paths) != str:
        raise ValueError("Invalid binary_paths for 'asis' artifact type. It must contain single item. %s items instead" % len(binary_paths))

    artifact_name = "%s%s%s" % (artifact_prefix, version, artifact_suffix,)
    return {
        'artifact_name': artifact_name,
        'artifact_path': os.path.join(root_dir, binary_paths[0])
    }


def create_artifact(root_dir, version, package, artifact_type, artifact_prefix, artifact_suffix, binary_paths):
    """Builds the artifact for upload"""
    logger.log_start("Creating artifact for package %s" % package)

    if artifact_type == 'zip':
        result = create_zip_artifact(root_dir, version, package, artifact_prefix, artifact_suffix, binary_paths)
    elif artifact_type == 'asis':
        result = create_asis_artifact(root_dir, version, package, artifact_prefix, artifact_suffix, binary_paths)
    else:
        raise ValueError("Invalid type specified; expected one of %s and got %s" % (AVAILABLE_ARTIFACT_TYPES, artifact_type,))

    return result


def check_version(version, build_version):
    """Fail deploy if tag version doesn't match version"""
    logger.log_start("Asserting versions match")
    if version != build_version:
        raise ValueError("Version extracted from build [%s] doesn't match declared in config [%s]" % (build_version, version))
    else:
        logger.log_info("Version match!")
    logger.log_done()


def execute_commands(commands):
    """Execute build commands"""
    logger.log_start("Executing build commands")
    for command in commands:
        utils.execute([command], shell=True)
    logger.log_done()



"""
    __main__.py

    Copyright (c) 2016 Snowplow Analytics Ltd. All rights reserved.

    This program is licensed to you under the Apache License Version 2.0,
    and you may not use this file except in compliance with the Apache License
    Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
    http://www.apache.org/licenses/LICENSE-2.0.

    Unless required by applicable law or agreed to in writing,
    software distributed under the Apache License Version 2.0 is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
    express or implied. See the Apache License Version 2.0 for the specific
    language governing permissions and limitations there under.

    Authors: Joshua Beemster
    Copyright: Copyright (c) 2016 Snowplow Analytics Ltd
    License: Apache License Version 2.0
"""

import argparse

import release_manager.targets.bintray as bintray
import release_manager.targets.awss3 as s3

import release_manager._version as _version
import release_manager.logger as logger
import release_manager.utils as utils
import release_manager.package as pack


# --- Main


def main():
    """Main function entry point"""
    parser = argparse.ArgumentParser(description="Utility for creating and uploading zip packages.")
    parser.add_argument("--config", help="the path to the configuration yaml file")
    parser.add_argument("--make-version", action='store_true', default=False, help="makes a new version for the package (bintray-specific)")
    parser.add_argument("--make-artifact", action='store_true', default=False, help="makes the artifacts that will be uploaded")
    parser.add_argument("--upload-artifact", action='store_true', default=False, help="uploads the artifacts to the targets")
    parser.add_argument("--check-version", action='store_true', default=False, help="checks that the version specified matches the build")
    parser.add_argument("--version", action='version', version=_version.__version__)
    args = parser.parse_args()

    # Parse args
    if not args.config:
        raise ValueError("A config must be passed to the program")
    config = utils.parse_config(args.config)

    if not args.make_version and not args.make_artifact and not args.upload_artifact and not args.check_version:
        logger.log_start("No actions selected, quitting")
        exit(0)

    if not args.make_artifact and args.upload_artifact:
        raise ValueError("Cannot upload artifact without first creating; please add '--make-artifact' to resolve...")

    logger.log_header("Starting Package uploader...")

    # Upload packages
    for package in config["packages"]:

        logger.log_start("Processing package %s" % package["name"])

        # Check the versions
        if args.check_version:
            pack.check_version(package["version"], package["build_version"])

        # Push to targets
        for target in config['targets']:
            if target['type'] == 'bintray':
                bintray.deploy_to_bintray(args, config['local'], package, target)
            elif target['type'] == 'awss3':
                s3.upload_to_s3(args, config['local'], package, target)
            else:
                raise ValueError("Invalid target specified; expected one of [bintray, awss3] and got %s" % target["type"])

        logger.log_footer("Finished processing package %s!" % package["name"])


if __name__ == "__main__":
    main()



"""
    logger.py

    Copyright (c) 2016 Snowplow Analytics Ltd. All rights reserved.

    This program is licensed to you under the Apache License Version 2.0,
    and you may not use this file except in compliance with the Apache License
    Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
    http://www.apache.org/licenses/LICENSE-2.0.

    Unless required by applicable law or agreed to in writing,
    software distributed under the Apache License Version 2.0 is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
    express or implied. See the Apache License Version 2.0 for the specific
    language governing permissions and limitations there under.

    Authors: Joshua Beemster
    Copyright: Copyright (c) 2016 Snowplow Analytics Ltd
    License: Apache License Version 2.0
"""


from __future__ import print_function


def log_header(header):
    """Styling for a function header log statement"""
    print("===============================================================")
    print(" %s" % header)
    print("---------------------------------------------------------------")


def log_footer(footer):
    """Styling for a function footer log statement"""
    print("---------------------------------------------------------------")
    print(" %s" % footer)
    print("===============================================================")


def log_start(start):
    """Styling for a block start log statement"""
    print("%s...\n" % start)


def log_done():
    """Styling for a block done log statement"""
    print(" - Done!\n")


def log_info(info):
    """Styling for a block info log statement"""
    if info != "":
        print(" + %s" % info)


def log_output(output):
    """Styling for an output log statement"""
    if output != "":
        for line in output.splitlines():
            print("   - %s" % line)



"""release_manager: __init__.py declaration"""

import release_manager.targets
import release_manager.__main__
import release_manager._version
import release_manager.logger
import release_manager.package
import release_manager.utils
